import requests
from bs4 import BeautifulSoup, PageElement
from luxhouse.database import SQLite

from luxhouse.model import House, Location


def load_page(url: str) -> BeautifulSoup:
    print('loading: ' + url)
    r = requests.get(url)
    print(r.status_code)
    return BeautifulSoup(r.content, 'html5lib')


def get_next_page(page: BeautifulSoup) -> PageElement:
    next = page.find_all('a', attrs={'class', 'in-pagination__item'})
    for candidate in next:
        label = candidate.find('span', attrs={'class', 'in-pagination__controlText'})
        if label and label.getText() == 'Suivant':
            return load_page(candidate['href'])
    
    return None


def to_int(page_element:PageElement)-> int:
    text = to_text(page_element)

    if text is None:
        return None

    elements = text.split('-')
    
    if len(elements) == 2:
        text = elements[0]

    digits = [i for i in text if i.isdecimal()]

    if len(digits) == 0:
        return None

    return int(''.join(digits))

def to_text(page_element: PageElement) -> str:
    if page_element is None:
        return None

    return page_element.get_text()

def extract_locality(title: str) -> tuple[str,str]:
    elements = title.split(',')

    if len(elements) >= 3:
        return elements[-2].strip(), elements[-1].strip()

    if len(elements) == 2:
        return elements[-1].strip(), elements[-1].strip()

    elements = title.split('à')
    
    if len(elements) == 2:
        return elements[-1].strip(), elements[-1].strip()

    raise TypeError('Failed to parse title ' + title)


def process_house(result: PageElement) -> House:
    header = result.find('a', attrs={'class': 'in-card__title'})
    reference = header['href']

    locality, commune = extract_locality(header['title'])
    location = Location(locality, commune, 'Luxembourg')
    title = header['title']
    current_price = to_int(result.find('li', attrs={'class':'in-realEstateListCard__features--main'}))
    rooms = to_int(result.find('li', attrs={'aria-label': 'chambres/pièces'}))
    bathrooms = to_int(result.find('li', attrs={'aria-label': 'salles de bain'}))
    size = to_int(result.find('li', attrs={'aria-label': 'superficie'}))
    description = to_text(result.find('p', attrs={'class': 'in-realEstateListCard__description'}))
    return House(reference, title, current_price, rooms, bathrooms, size, location, description)
        
def result_reader(url: str):
    page = load_page(url)

    while page:
        results = page.find('ul', attrs={'data-cy': 'result-list'})
        for result in results.find_all('div', attrs={'class': 'in-realEstateListCard__content'}):
            yield result
        page = get_next_page(page)


def immotop(database_path:str) -> None:
    URL = "https://www.immotop.lu/vente-maisons/luxembourg-pays/?criterio=dataModifica&ordine=desc&noAste=1"

    database = SQLite(database_path)

    for result in result_reader(URL):
        house = process_house(result)
        database.add_house(house)

if __name__ == '__main__':
    immotop('results.sqlite3')


from datetime import date
from sqlite3 import Cursor

def fetchOne(cursor: Cursor, sql:str, values:tuple) -> any:
    cursor.execute(sql, values)
    records = cursor.fetchone()
    
    if records is None:
        with_values = ''
        for item in values:
            with_values = with_values + str(item) + ','
        
        if with_values:
            with_values = ' with values (' + with_values + ')'

        raise Exception("query found nothin: " + sql + with_values)

    return records[0]

class PriceDate(object):
    def __init__(self, price:int, date:date, house_id:int) -> None:
        self.price = price
        self.date = date
        self.house_id = house_id

    def create_table() -> str:
        return """
            create table if not exists price_date(
                id integer primary key,
                price integer,
                date date not null,
                house_id integer not null,
                unique(price, date, house_id),
                constraint fk_house foreign key (house_id) references houses (id)
            );"""
    
    def insert(self, cursor:Cursor):
        if not self.is_saved(cursor):
            sql = "insert into price_date(price,date,house_id) values (?,?,?);"
            values = (self.price, self.date, self.house_id)
            cursor.execute(sql, values)
        
        self.id = self.get_id(cursor)
        return self

    def get_id(self, cursor: Cursor) -> int:
        sql = "select id from price_date where (price=? or price is null) and date=? and house_id=?;"
        values = (self.price, self.date, self.house_id)
        return fetchOne(cursor, sql, values)

    def is_saved(self, cursor: Cursor) -> bool:
        try:
            self.get_id(cursor)
            return True
        except:
            return False

class Location(object):
    def __init__(self, locality:str, commune:str, country:str) -> None:
        self.locality = locality
        self.commune = commune
        self.country = country

    def create_table() -> str:
        return """
            create table if not exists locations(
                id integer primary key,
                locality text not null,
                commune text not null,
                country text not null,
                unique(locality,commune, country)
            );"""

    def insert(self, cursor: Cursor):
        if not self.is_saved(cursor):
            sql = "insert into locations(locality,commune,country) values (?,?,?);"
            values = (self.locality, self.commune, self.country)
            cursor.execute(sql, values)
        
        self.id = self.get_id(cursor)
        return self
    
    def get_id(self, cursor: Cursor) -> int:
        sql = "select id from locations where locality=? and commune=? and country=?;"
        values = (self.locality, self.commune, self.country)
        return fetchOne(cursor, sql, values)

    def is_saved(self, cursor: Cursor) -> bool:
        try:
            self.get_id(cursor)
            return True
        except:
            return False


class House(object):
    def __init__(self, reference:str, title:str, current_price:int, rooms:int, bathrooms:int, size:int, location: Location, description:str) -> None:
        self.reference = reference
        self.title = title
        self.rooms = rooms 
        self.bathrooms = bathrooms
        self.size = size
        self.location = location
        self.description = description
        self.current_date = date.today()
        self.current_price = current_price

    def create_table() -> str:
        return """
                create table if not exists houses(
                    id integer primary key,
                    reference text unique,
                    title text not null,
                    rooms integer,
                    bathrooms integer,
                    size integer,
                    location_id integer not null,
                    description text,
                    constraint fk_location foreign key (location_id) references locations (id)
                );"""

    def insert(self, cursor: Cursor):
        try:
            cursor.execute("begin")
            if not self.is_saved(cursor):
                location = self.location.insert(cursor)
                sql = "insert into houses(reference, title, rooms, bathrooms, size, location_id, description) values (?,?,?,?,?,?,?);"
                values = (self.reference, self.title, self.rooms, self.bathrooms, self.size, location.id, self.description)
                cursor.execute(sql, values)
                cursor.execute("commit")
            self.id = self.get_id(cursor)
            price_date = PriceDate(self.current_price, self.current_date, self.id)
            price_date.insert(cursor)
        except Exception as e:
            print("Failed to save house " + self.reference + ": " + str(e))
            cursor.execute('rollback')

        return self

    def get_id(self, cursor: Cursor) -> int:
        sql = "select id from houses where reference=?;"
        values = (self.reference,)
        return fetchOne(cursor, sql, values)

    def is_saved(self, cursor: Cursor) -> bool:
        try:
            self.get_id(cursor)
            return True
        except:
            return False
